from contribution.models import Premium
from insuree.models import InsureePolicy, Family, Gender, Insuree
from insuree.test_helpers import create_test_insuree
from policy.models import Policy
from policy.values import policy_values
from product.models import Product
from core.utils import filter_validity
import datetime


def create_test_policy(product, insuree, link=True, valid=True, custom_props=None, check=False):
    """
    Compatibility method that only return the Policy
    """
    return create_test_policy2(product, insuree, link, valid, custom_props, check)[0]

def dts(s):
    return datetime.datetime.strptime(s, "%Y-%m-%d").date()

def create_test_policy2(product, insuree, link=True, valid=True, custom_props=None, check=False):
    """
    Creates a Policy and optionally an InsureePolicy
    :param product: Product on which this Policy is based (or its ID)
    :param insuree: The Policy will be linked to this Insuree's family and if link is True, the InsureePolicy will
     belong to him
    :param link: if True (default), an InsureePolicy will also be created
    :param valid: Whether the created Policy should be valid (validity_to)
    :param custom_props: dictionary of custom values for the Policy, when overriding a foreign key, override the _id
    :return: The created Policy and InsureePolicy
    """
    policy = Policy.objects.create(
        **{
            "family": insuree.family,
            "product_id": product.id if isinstance(product, Product) else product,
            "status": Policy.STATUS_ACTIVE,
            "stage": Policy.STAGE_NEW,
            "enroll_date": dts("2019-01-01"),
            "start_date": dts("2019-01-02"),
            "validity_from": dts("2019-01-01"),
            "effective_date": dts("2019-01-01"),
            "expiry_date": dts("2039-06-01"),
            "validity_to": None if valid else dts("2019-01-01"),
            "audit_user_id": -1,
            **(custom_props if custom_props else {})
        }
    )
    if link:
        insuree_policy = InsureePolicy.objects.create(
            insuree=insuree,
            policy=policy,
            audit_user_id=-1,
            effective_date=dts("2019-01-01"),
            expiry_date=dts("2039-06-01"),
            validity_from=dts("2019-01-01"),
            validity_to=None if valid else dts("2019-01-01"),
        )
    else:
        insuree_policy = None

    # Was added for OMT-333 but breaks tests that explicitly call policy_values
    policy, warnings = policy_values(policy, insuree.family, None)
    if check and warnings:
        raise Exception("Policy has warnings: {}".format(warnings))
    return policy, insuree_policy


def create_test_policy_with_IPs(product, insuree, valid=True, policy_props=None, IP_props=None, premium_props=None):
    """
    Creates a Policy with its related InsureePolicys for each Family member.
    :param product: Product on which this Policy is based (or its ID)
    :param insuree: The Policy will be linked to this Insuree's family
    :param valid: Whether the created Policy should be valid (validity_to)
    :param policy_props: dictionary of custom values for the Policy, when overriding a foreign key, override the _id
    :param IP_props: dictionary of custom values for the InsureePolicy, when overriding a foreign key, override the _id
    :param premium_props: dictionary of custom values for the Premium, when overriding a foreign key, override the _id
    :return: The created Policy and InsureePolicy
    """
    value = 10000.00
    default_date = dts("2019-01-01")
    start_date = dts("2019-01-02")
    expiry_date = dts("2039-06-01")

    policy = Policy.objects.create(
        **{
            "family": insuree.family,
            "product_id": product.id if isinstance(product, Product) else product,
            "status": Policy.STATUS_ACTIVE,
            "stage": Policy.STAGE_NEW,
            "enroll_date": default_date,
            "start_date": start_date,
            "validity_from": default_date,
            "effective_date": default_date,
            "expiry_date": expiry_date,
            "value": value,
            "validity_to": None if valid else default_date,
            "audit_user_id": -1,
            **(policy_props if policy_props else {})
        }
    )

    for member in policy.family.members.filter(validity_to__isnull=True):
        InsureePolicy.objects.create(
            **{
                "insuree": member,
                "policy": policy,
                "audit_user_id": -1,
                "effective_date": default_date,
                "expiry_date": expiry_date,
                "start_date": start_date,
                "enrollment_date": default_date,
                "validity_from": default_date,
                "validity_to": None if valid else default_date,
                **(IP_props if IP_props else {})
            }
        )

    Premium.objects.create(
        **{
            "amount": value,
            "receipt": "TEST1234",
            "pay_date": default_date,
            "pay_type": "C",
            "validity_from": default_date,
            "policy_id": policy.id,
            "audit_user_id": -1,
            "validity_to": None if valid else default_date,
            "created_date": default_date,
            **(premium_props if premium_props else {})
        }
    )

    return policy


def create_test_insuree_for_policy(with_family=True, is_head=False, custom_props=None, family_custom_props=None):
    # To establish the mandatory reference between "Insuree" and "Family" objects, we can insert the "Family" object
    # with a temporary ID and later update it to associate with the respective "Insuree" object.
    insuree=  create_test_insuree(with_family=with_family, is_head=is_head, custom_props=custom_props, family_custom_props=family_custom_props)


    return insuree, insuree.family



import logging

import graphene
from django.db import transaction

from policy.services import PolicyService

from .apps import PolicyConfig
from core.schema import OpenIMISMutation
from .models import Policy, PolicyMutation
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ValidationError, PermissionDenied
from django.utils.translation import gettext as _
from .validations import validate_idle_policy

logger = logging.getLogger(__name__)


class PolicyInputType(OpenIMISMutation.Input):
    # several fields (such as status, stage,...) are managed "internally"
    # and only initialized/updated via dedicated mutations (renew , cancel,...)
    id = graphene.Int(required=False, read_only=True)
    uuid = graphene.String(required=False)
    enroll_date = graphene.Date(required=True)
    start_date = graphene.Date(required=True)
    expiry_date = graphene.Date(required=True)
    value = graphene.Decimal(max_digits=18, decimal_places=2, required=True)
    product_id = graphene.Int(required=True)
    family_id = graphene.Int(required=True)
    officer_id = graphene.Int(required=True)
    is_paid = graphene.Boolean(required=False)
    receipt = graphene.String(required=False)
    payer_uuid = graphene.String(required=False)


class CreateRenewOrUpdatePolicyMutation(OpenIMISMutation):
    @classmethod
    def do_mutate(cls, perms, user, **data):
        if type(user) is AnonymousUser or not user.id:
            raise ValidationError(
                _("mutation.authentication_required"))
        if not user.has_perms(perms):
            raise PermissionDenied(_("unauthorized"))
        client_mutation_id = data.get("client_mutation_id")
        errors = validate_idle_policy(data)
        if len(errors):
            return errors
        data['audit_user_id'] = user.id_for_audit
        from core.utils import TimeUtils
        data['validity_from'] = TimeUtils.now()
        policy = PolicyService(user).update_or_create(data, user)
        PolicyMutation.object_mutated(user, client_mutation_id=client_mutation_id, policy=policy)
        return None


class CreatePolicyMutation(CreateRenewOrUpdatePolicyMutation):
    _mutation_module = "policy"
    _mutation_class = "CreatePolicyMutation"

    class Input(PolicyInputType):
        pass

    @classmethod
    def async_mutate(cls, user, **data):
        try:
            with transaction.atomic():
                data["status"] = Policy.STATUS_IDLE
                data["stage"] = Policy.STAGE_NEW
                return cls.do_mutate(PolicyConfig.gql_mutation_create_policies_perms, user, **data)
        except Exception as exc:
            return [{
                'message': _("policy.mutation.failed_to_create_policy"),
                'detail': str(exc),
                'exc': exc}]


class UpdatePolicyMutation(CreateRenewOrUpdatePolicyMutation):
    _mutation_module = "policy"
    _mutation_class = "UpdatePolicyMutation"

    class Input(PolicyInputType):
        pass

    @classmethod
    def async_mutate(cls, user, **data):
        try:
            with transaction.atomic():
                return cls.do_mutate(PolicyConfig.gql_mutation_edit_policies_perms, user, **data)
        except Exception as exc:
            return [{
                'message': _("policy.mutation.failed_to_update_policy"),
                'detail': str(exc),
                'exc': exc}]


class RenewPolicyMutation(CreateRenewOrUpdatePolicyMutation):
    _mutation_module = "policy"
    _mutation_class = "RenewPolicyMutation"

    class Input(PolicyInputType):
        pass

    @classmethod
    def async_mutate(cls, user, **data):
        try:
            with transaction.atomic():
                # ensure we don't update the existing one, but recreate a new one!
                if 'policy_uuid' in data:
                    data.pop('policy_uuid')
                data["status"] = Policy.STATUS_IDLE
                data["stage"] = Policy.STAGE_RENEWED
                return cls.do_mutate(PolicyConfig.gql_mutation_renew_policies_perms, user, **data)
        except Exception as exc:
            return [{
                'message': _("policy.mutation.failed_to_renew_policy"),
                'detail': str(exc),
                'exc': exc}]


class SuspendPoliciesMutation(OpenIMISMutation):
    _mutation_module = "policy"
    _mutation_class = "SuspendPolicyMutation"

    class Input(OpenIMISMutation.Input):
        uuids = graphene.List(graphene.String)

    @classmethod
    def async_mutate(cls, user, **data):
        try:
            with transaction.atomic():
                if type(user) is AnonymousUser or not user.id:
                    raise ValidationError(
                        _("mutation.authentication_required"))
                if not user.has_perms(PolicyConfig.gql_mutation_suspend_policies_perms):
                    raise PermissionDenied(_("unauthorized"))
                errors = []
                for policy_uuid in data["uuids"]:
                    policy = Policy.objects.filter(uuid=policy_uuid).first()
                    if policy is None:
                        errors += {
                            'title': policy_uuid,
                            'list': [{'message': _(
                                "policy.mutation.id_does_not_exist") % {'id': policy_uuid}}]
                        }
                        continue
                    errors += PolicyService(user).set_suspended(user, policy)
                if len(errors) == 1:
                    errors = errors[0]['list']
                return errors
        except Exception as exc:
            return [{
                'message': _("policy.mutation.failed_to_suspend_policy"),
                'detail': str(exc),
                'exc': exc}]


class DeletePoliciesMutation(OpenIMISMutation):
    _mutation_module = "policy"
    _mutation_class = "DeletePoliciesMutation"

    class Input(OpenIMISMutation.Input):
        uuids = graphene.List(graphene.String)

    @classmethod
    def async_mutate(cls, user, **data):
        try:
            with transaction.atomic():
                if not user.has_perms(PolicyConfig.gql_mutation_delete_policies_perms):
                    raise PermissionDenied(_("unauthorized"))
                errors = []
                for policy_uuid in data["uuids"]:
                    policy = Policy.objects \
                        .filter(uuid=policy_uuid) \
                        .first()
                    if policy is None:
                        errors += {
                            'title': policy_uuid,
                            'list': [{'message': _(
                                "policy.validation.id_does_not_exist") % {'id': policy_uuid}}]
                        }
                        continue
                    errors += PolicyService(user).set_deleted(policy)
                if len(errors) == 1:
                    errors = errors[0]['list']
                return errors
        except Exception as exc:
            return [{
                'message': _("policy.mutation.failed_to_delete_policies"),
                'detail': str(exc),
                'exc': exc}]



import graphene
from graphene_django import DjangoObjectType
from django.utils.translation import gettext as _
from .apps import PolicyConfig
from .models import Policy
from core import prefix_filterset, filter_validity, ExtendedConnection, ExtendedRelayConnection
from core.schema import OfficerGQLType
from product.schema import ProductGQLType
from django.core.exceptions import PermissionDenied


class PolicyGQLType(DjangoObjectType):
    sum_premiums = graphene.Float(source="sum_premiums")
    
    def resolve_family(self, info):
        if not info.context.user.has_perms(PolicyConfig.gql_query_policies_perms):
            raise PermissionDenied(_("unauthorized"))
        if "family_loader" in info.context.dataloaders and self.family_id:
            return info.context.dataloaders["family_loader"].load(self.family_id)
        return self.family

    def resolve_product(self, info):
        if not info.context.user.has_perms(PolicyConfig.gql_query_policies_perms):
            raise PermissionDenied(_("unauthorized"))
        if "product_loader" in info.context.dataloaders and self.product_id:
            return info.context.dataloaders["product_loader"].load(self.product_id)
        return self.product

    class Meta:
        model = Policy
        interfaces = (graphene.relay.Node,)
        filter_fields = {
            "id": ["exact"],
            "uuid": ["exact"],
            "enroll_date": ["exact", "lt", "lte", "gt", "gte"],
            "start_date": ["exact", "lt", "lte", "gt", "gte"],
            "effective_date": ["exact", "lt", "lte", "gt", "gte"],
            "expiry_date": ["exact", "lt", "lte", "gt", "gte"],
            "stage": ["exact"],
            "status":  ["exact", "lt", "lte", "gt", "gte"],
            "value": ["exact", "lt", "lte", "gt", "gte"],
            **prefix_filterset("product__", ProductGQLType._meta.filter_fields),
            **prefix_filterset("officer__", OfficerGQLType._meta.filter_fields),
        }
        connection_class = ExtendedConnection


class PolicyAndWarningsGQLType(graphene.ObjectType):
    policy = graphene.Field(PolicyGQLType)
    warnings = graphene.List(graphene.String)


class PolicyByFamilyOrInsureeGQLType(graphene.ObjectType):
    class Meta:
        interfaces = (graphene.relay.Node,)

    policy_id = graphene.Int()
    policy_uuid = graphene.String()
    policy_value = graphene.Float()
    product_code = graphene.String()
    product_name = graphene.String()
    start_date = graphene.Date()
    enroll_date = graphene.Date()
    effective_date = graphene.Date()
    expiry_date = graphene.Date()
    officer_code = graphene.String()
    officer_name = graphene.String()
    status = graphene.Int()
    ded = graphene.Float()
    ded_in_patient = graphene.Float()
    ded_out_patient = graphene.Float()
    ceiling = graphene.Float()
    ceiling_in_patient = graphene.Float()
    ceiling_out_patient = graphene.Float()
    balance = graphene.Float()
    validity_from = graphene.Date()
    validity_to = graphene.Date()
    max_installments = graphene.Int()


class PolicyByFamilyOrInsureeConnection(ExtendedRelayConnection):
    class Meta:
        node = PolicyByFamilyOrInsureeGQLType


class EligibilityGQLType(graphene.ObjectType):
    prod_id = graphene.String()
    total_admissions_left = graphene.Int()
    total_visits_left = graphene.Int()
    total_consultations_left = graphene.Int()
    total_surgeries_left = graphene.Int()
    total_deliveries_left = graphene.Int()
    total_antenatal_left = graphene.Int()
    consultation_amount_left = graphene.Float()
    surgery_amount_left = graphene.Float()
    delivery_amount_left = graphene.Float()
    hospitalization_amount_left = graphene.Float()
    antenatal_amount_left = graphene.Float()
    min_date_service = graphene.types.datetime.Date()
    min_date_item = graphene.types.datetime.Date()
    service_left = graphene.Int()
    item_left = graphene.Int()
    is_item_ok = graphene.Boolean()
    is_service_ok = graphene.Boolean()




from unittest import mock, skip

from claim.test_helpers import create_test_claim, create_test_claimservice, create_test_claimitem
from claim.validations import validate_claim, validate_assign_prod_to_claimitems_and_services, process_dedrem
from core.models import InteractiveUser, User
from core.test_helpers import create_test_officer
from django.conf import settings
from django.test import TestCase
from insuree.test_helpers import create_test_photo
from medical.test_helpers import create_test_item, create_test_service
from medical_pricelist.test_helpers import add_service_to_hf_pricelist, add_item_to_hf_pricelist
from insuree.test_helpers import create_test_insuree
from policy.test_helpers import create_test_policy2, create_test_insuree_for_policy
from product.test_helpers import create_test_product, create_test_product_service, create_test_product_item
from location.test_helpers import create_test_health_facility
from policy.services import *
from medical_pricelist.test_helpers import (
    create_test_item_pricelist,
    create_test_service_pricelist
)

class EligibilityServiceTestCase(TestCase):
    def setUp(self) -> None:
        super(EligibilityServiceTestCase, self).setUp()
        self.user = mock.Mock(is_anonymous=False)
        self.user.has_perms = mock.MagicMock(return_value=True)

    def test_eligibility_request_permission_denied(self):
        with mock.patch("django.db.backends.utils.CursorWrapper") as mock_cursor:
            mock_cursor.return_value.description = None
            mock_user = mock.Mock(is_anonymous=False)
            mock_user.has_perms = mock.MagicMock(return_value=False)
            req = EligibilityRequest(chf_id="a")
            service = EligibilityService(mock_user)
            with self.assertRaises(PermissionDenied) as cm:
                service.request(req)
            mock_user.has_perms.assert_called_with(PolicyConfig.gql_query_eligibilities_perms)

    @skip("this test hangs on psql, the mock destroys normal queries happening inside EligibilityRequest")
    def test_eligibility_request_all_good(self):
        with mock.patch("django.db.backends.utils.CursorWrapper") as mock_cursor:
            return_values = [
                list(range(1, 13)),
                [
                    core.datetime.date(2020, 1, 9),
                    core.datetime.date(2020, 1, 10),
                    20,
                    21,
                    True,
                    True,
                ],
            ]
            # required for all modules tests
            mock_cursor.return_value.fetchone.side_effect = return_values
            # required for policy module tests
            mock_cursor.return_value.__enter__.return_value.fetchone.side_effect = return_values
            mock_user = mock.Mock(is_anonymous=False)
            insuree, family = create_test_insuree_for_policy(custom_props={"chf_id": "tier1234" })
            product = create_test_product("ELI1")
            create_test_policy2(product, insuree)
            mock_user.has_perm = mock.MagicMock(return_value=True)
            req = EligibilityRequest(chf_id="tier1234")
            service = EligibilityService(mock_user)
            res = service.request(req)

            expected = EligibilityResponse(
                eligibility_request=req,
                prod_id=product.id,
                total_admissions_left=None,
                total_visits_left=None,
                total_consultations_left=None,
                total_surgeries_left=None,
                total_deliveries_left=None,
                total_antenatal_left=None,
                consultation_amount_left=None,
                surgery_amount_left=None,
                delivery_amount_left=None,
                hospitalization_amount_left=None,
                antenatal_amount_left=None,
                min_date_service=None,
                min_date_item=None,
                service_left=None,
                item_left=None,
                is_item_ok=True,
                is_service_ok=True,
            )
            self.assertEquals(expected, res)

    def test_eligibility_sp_call(self):
        if not connection.vendor == "mssql":
            self.skipTest("This test can only be executed for MSSQL database")
        mock_user = mock.Mock(is_anonymous=False)
        mock_user.has_perm = mock.MagicMock(return_value=True)
        req = EligibilityRequest(chf_id="070707070")
        service = StoredProcEligibilityService(mock_user)
        res = service.request(req, EligibilityResponse(req))
        expected = EligibilityResponse(
            eligibility_request=req,
            prod_id=4,
            total_admissions_left=None,
            total_visits_left=None,
            total_consultations_left=None,
            total_surgeries_left=None,
            total_deliveries_left=None,
            total_antenatal_left=None,
            consultation_amount_left=None,
            surgery_amount_left=None,
            delivery_amount_left=None,
            hospitalization_amount_left=None,
            antenatal_amount_left=None,
            min_date_service=None,
            min_date_item=None,
            service_left=None,
            item_left=None,
            is_item_ok=True,
            is_service_ok=True,
        )
        self.assertEquals(expected, res)

    def test_eligibility_stored_proc_serv(self):
        if not connection.vendor == "mssql":
            self.skipTest("This test can only be executed for MSSQL database")
        for category in [
            Service.CATEGORY_SURGERY,
            Service.CATEGORY_CONSULTATION,
            Service.CATEGORY_HOSPITALIZATION,
            Service.CATEGORY_OTHER,
            Service.CATEGORY_ANTENATAL,
        ]:
            with self.subTest(category=category):
                self.eligibility_serv(category)

    def eligibility_serv(self, category):
        insuree, family = create_test_insuree_for_policy(custom_props={"chf_id": "elgsp" + category})
        product = create_test_product("ELI1")
        (policy, insuree_policy) = create_test_policy2(product, insuree)
        service = create_test_service(category)
        svc_pl_detail = add_service_to_hf_pricelist(service)
        product_service = create_test_product_service(product, service, custom_props={"limit_no_adult": 20})
        claim = create_test_claim(custom_props={"insuree_id": insuree.id})
        claim_service = create_test_claimservice(claim, custom_props={"service_id": service.id})
        errors = validate_claim(claim, True)
        errors += validate_assign_prod_to_claimitems_and_services(claim)
        errors += process_dedrem(claim, -1, True)
        self.assertEqual(len(errors), 0)

        native_el_svc = NativeEligibilityService(self.user)
        req = EligibilityRequest(chf_id=insuree.chf_id, service_code=service.code)
        expected_resposnse = EligibilityResponse(
            antenatal_amount_left=None,
            consultation_amount_left=None,
            delivery_amount_left=None,
            eligibility_request=req,
            final=False,
            hospitalization_amount_left=None,
            is_item_ok=True,
            is_service_ok=True,
            item_left=None,
            min_date_item=core.datetime.date(2019, 3, 1),
            min_date_service=None,
            prod_id=product.id,
            service_left=13.00,
            surgery_amount_left=None,
            total_visits_left=None,
            total_antenatal_left=None,
            total_surgeries_left=None,
            total_admissions_left=None,
            total_deliveries_left=None,
            total_consultations_left=None,
        )
        settings.ROW_SECURITY = False
        native_response = native_el_svc.request(req, EligibilityResponse(req))
        self.assertIsNotNone(native_response)
        self.assertEquals(native_response, expected_resposnse)

        family.delete()
        claim.dedrems.all().delete()
        claim_service.delete()
        claim.delete()
        product_service.delete()
        svc_pl_detail.delete()
        service.delete()
        policy.insuree_policies.all().delete()
        policy.delete()
        product.delete()
        insuree.delete()

    def test_eligibility_item(self):
        if not connection.vendor == "mssql":
            self.skipTest("This test can only be executed for MSSQL database")
        insuree, family = create_test_insuree_for_policy()
        product = create_test_product("ELI1")
        (policy, insuree_policy) = create_test_policy2(product, insuree)
        item = create_test_item("A")
        item_pl_detail = add_item_to_hf_pricelist(item)
        product_item = create_test_product_item(product, item, custom_props={"limit_no_adult": 12})
        claim = create_test_claim(custom_props={"insuree_id": insuree.id})
        claim_item = create_test_claimitem(claim, "A", custom_props={"item_id": item.id})
        errors = validate_claim(claim, True)
        errors += validate_assign_prod_to_claimitems_and_services(claim)
        errors += process_dedrem(claim, -1, True)
        self.assertEqual(len(errors), 0)

        native_el_svc = NativeEligibilityService(self.user)
        req = EligibilityRequest(chf_id=insuree.chf_id, item_code=item.code)
        expected_resposnse = EligibilityResponse(
            antenatal_amount_left=None,
            consultation_amount_left=None,
            delivery_amount_left=None,
            eligibility_request=req,
            final=False,
            hospitalization_amount_left=None,
            is_item_ok=True,
            is_service_ok=True,
            item_left=5.00,
            min_date_item=core.datetime.date(2019, 3, 1),
            min_date_service=None,
            prod_id=product.id,
            service_left=None,
            surgery_amount_left=None,
            total_visits_left=None,
            total_antenatal_left=None,
            total_surgeries_left=None,
            total_admissions_left=None,
            total_deliveries_left=None,
            total_consultations_left=None,
        )
        settings.ROW_SECURITY = False
        native_response = EligibilityResponse(req)
        native_response = native_el_svc.request(req, native_response)
        self.assertIsNotNone(native_response)
        self.assertEquals(native_response, expected_resposnse)

        family.delete()
        claim.dedrems.all().delete()
        claim_item.delete()
        claim.delete()
        product_item.delete()
        item_pl_detail.delete()
        item.delete()
        policy.insuree_policies.all().delete()
        policy.delete()
        product.delete()
        insuree.delete()

    def test_eligibility_by_insuree(self):
        if not connection.vendor == "mssql":
            self.skipTest("This test can only be executed for MSSQL database")
        insuree, family = create_test_insuree_for_policy()
        product = create_test_product("ELI1")
        (policy, insuree_policy) = create_test_policy2(product, insuree)
        item = create_test_item("A")
        item_pl_detail = add_item_to_hf_pricelist(item)
        product_item = create_test_product_item(product, item, custom_props={"limit_no_adult": 12})
        claim = create_test_claim(custom_props={"insuree_id": insuree.id})
        claim_item = create_test_claimitem(claim, "A", custom_props={"item_id": item.id})
        errors = validate_claim(claim, True)
        errors += validate_assign_prod_to_claimitems_and_services(claim)
        errors += process_dedrem(claim, -1, True)
        self.assertEqual(len(errors), 0)

        native_el_svc = NativeEligibilityService(self.user)
        req = EligibilityRequest(chf_id=insuree.chf_id, item_code=item.code)
        expected_resposnse = EligibilityResponse(
            antenatal_amount_left=None,
            consultation_amount_left=None,
            delivery_amount_left=None,
            eligibility_request=req,
            final=False,
            hospitalization_amount_left=None,
            is_item_ok=True,
            is_service_ok=True,
            item_left=5.00,
            min_date_item=core.datetime.date(2019, 3, 1),
            min_date_service=None,
            prod_id=product.id,
            service_left=None,
            surgery_amount_left=None,
            total_visits_left=None,
            total_antenatal_left=None,
            total_surgeries_left=None,
            total_admissions_left=None,
            total_deliveries_left=None,
            total_consultations_left=None,
        )
        settings.ROW_SECURITY = False
        native_response = EligibilityResponse(req)
        native_response = native_el_svc.request(req, native_response)
        self.assertIsNotNone(native_response)
        self.assertEquals(native_response, expected_resposnse)

        family.delete()
        claim.dedrems.all().delete()
        claim_item.delete()
        claim.delete()
        product_item.delete()
        item_pl_detail.delete()
        item.delete()
        policy.insuree_policies.all().delete()
        policy.delete()
        product.delete()
        insuree.delete()

    @skip("Not sure what is the proper behaviour when an IP is not present, skipping for now so that the main case"
          "can be fixed.")
    def test_eligibility_stored_proc_item_no_insuree_policy(self):
        insuree = create_test_insuree_for_policy()
        product = create_test_product("ELI1")
        (policy, _) = create_test_policy2(
            product, insuree, link=False, custom_props={"status": Policy.STATUS_IDLE})
        item = create_test_item("A")
        item_pl_detail = add_item_to_hf_pricelist(item)
        product_item = create_test_product_item(product, item, custom_props={"limit_no_adult": 12})

        sp_el_svc = StoredProcEligibilityService(self.user)
        native_el_svc = NativeEligibilityService(self.user)
        req = EligibilityRequest(chf_id=insuree.chf_id, item_code=item.code)
        settings.ROW_SECURITY = False
        sp_response = EligibilityResponse(req)
        sp_response = sp_el_svc.request(req, sp_response)
        native_response = EligibilityResponse(req)
        native_response = native_el_svc.request(req, native_response)
        self.assertIsNotNone(native_response)
        self.assertIsNotNone(sp_response)
        self.assertEquals(native_response, sp_response)

        product_item.delete()
        item_pl_detail.delete()
        item.delete()
        policy.insuree_policies.all().delete()
        policy.delete()
        product.delete()
        insuree.delete()

    def test_eligibility_signal(self):
        
        insuree, family = create_test_insuree_for_policy()
        #spl = create_test_service_pricelist(location_id=family.location.parent.id)
        #ipl = create_test_item_pricelist(location_id=family.location.parent.id)
        #hf =create_test_health_facility(code= 'tst-18', location_id=family.location.parent.id,  custom_props={'id':18, 'items_pricelist': ipl, 'services_pricelist': spl })

        product = create_test_product("ELI1")
        (policy, insuree_policy) = create_test_policy2(product, insuree)
        item = create_test_item("A")
        item_pl_detail = add_item_to_hf_pricelist(item)
        product_item = create_test_product_item(product, item, custom_props={"limit_no_adult": 12})
        claim = create_test_claim(custom_props={"insuree_id": insuree.id})
        claim_item = create_test_claimitem(claim, "A", custom_props={"item_id": item.id})
        errors = validate_claim(claim, True)
        errors += validate_assign_prod_to_claimitems_and_services(claim)
        errors += process_dedrem(claim, -1, True)
        self.assertEqual(len(errors), 0)

        def signal_before(sender, **kwargs):
            kwargs["response"].final = True
            kwargs["response"].total_admissions_left = 444719
            return kwargs["response"]

        signal_eligibility_service_before.connect(signal_before)

        el_svc = EligibilityService(self.user)
        req = EligibilityRequest(chf_id=insuree.chf_id, item_code=item.code)
        settings.ROW_SECURITY = False

        response = el_svc.request(req)
        self.assertIsNotNone(response)
        self.assertEquals(response.total_admissions_left, 444719)

        signal_eligibility_service_before.disconnect(signal_before)
        claim.dedrems.all().delete()
        claim_item.delete()
        claim.delete()
        product_item.delete()
        item_pl_detail.delete()
        item.delete()
        policy.insuree_policies.all().delete()
        policy.delete()
        product.delete()
        insuree.delete()
        family.delete()


class RenewalsTestCase(TestCase):
    item_1 = None

    def setUp(self) -> None:
        super(RenewalsTestCase, self).setUp()
        self.i_user = InteractiveUser(
            login_name="test_batch_run", audit_user_id=978911, id=97891
        )
        self.user = User(i_user=self.i_user)

        self.item_1 = create_test_item("D")

    def test_insert_renewals(self):
        # Given
        from core import datetime, datetimedelta

        insuree, family = create_test_insuree_for_policy()
        product = create_test_product("VISIT")
        officer = create_test_officer()

        (policy_not_expiring, inspolicy_not_expiring) = create_test_policy2(
            product=product,
            insuree=insuree,
            custom_props={"expiry_date": "2099-01-01", "officer": officer},
        )
        (policy_expiring, inspolicy_expiring) = create_test_policy2(
            product=product,
            insuree=insuree,
            custom_props={
                "expiry_date": datetime.datetime.now() + datetimedelta(days=5),
                "officer": officer,
            },
        )

        # when
        insert_renewals(officer_id=officer.id)

        # then
        renewals = PolicyRenewal.objects.filter(insuree=insuree)
        expected_renewal = renewals.filter(policy=policy_expiring).first()
        self.assertIsNotNone(expected_renewal)

        should_not_renew = renewals.filter(policy=policy_not_expiring).first()
        self.assertIsNone(should_not_renew)

        # tearDown
        renewals.delete()
        inspolicy_expiring.delete()
        policy_expiring.delete()
        inspolicy_not_expiring.delete()
        policy_not_expiring.delete()
        officer.delete()
        product.delete()
        insuree.delete()
        family.delete()

    def test_update_renewals(self):
        # Given
        from core import datetime, datetimedelta

        insuree, family = create_test_insuree_for_policy()
        product = create_test_product("VISIT")
        officer = create_test_officer()

        (policy_expiring, inspolicy_expiring) = create_test_policy2(
            product=product,
            insuree=insuree,
            custom_props={"expiry_date": "2019-01-01", "officer": officer},
        )
        (policy_not_expired_yet, inspolicy_not_expired_yet) = create_test_policy2(
            product=product,
            insuree=insuree,
            custom_props={
                "expiry_date": datetime.datetime.now() + datetimedelta(days=5),
                "officer": officer,
            },
        )

        # when
        update_renewals()

        # then
        policy_expiring.refresh_from_db()
        policy_not_expired_yet.refresh_from_db()

        self.assertEquals(policy_expiring.status, Policy.STATUS_EXPIRED)
        self.assertEquals(policy_not_expired_yet.status, Policy.STATUS_ACTIVE)

        # tearDown
        inspolicy_expiring.delete()
        policy_expiring.delete()
        inspolicy_not_expired_yet.delete()
        policy_not_expired_yet.delete()
        officer.delete()
        product.delete()
        insuree.delete()
        family.delete()

    def test_renewals_sms(self):
        # Given
        from core import datetime, datetimedelta

        insuree, family = create_test_insuree_for_policy(
            custom_props={"chf_id": "TESTCHFSMS", 'last_name':'Test Last',"phone": "+33644444719"}        )
        product = create_test_product("VISIT")
        officer = create_test_officer(
            custom_props={"phone": "+32444444444", "phone_communication": True},
            villages = [family.location]
        )

        (policy_expiring, _) = create_test_policy2(
            product=product,
            insuree=insuree,
            custom_props={"expiry_date": "2019-01-01", "officer": officer},
        )
        (policy_not_expired_yet, _) = create_test_policy2(
            product=product,
            insuree=insuree,
            custom_props={
                "expiry_date": datetime.datetime.now() + datetimedelta(days=5),
                "officer": officer,
            },
        )

        family_template = "FAMSMS;{{renewal.insuree.chf_id}};{{renewal.insuree.last_name}};{{renewal.new_product.name}}"

        insert_renewals(officer_id=officer.id)

        # when
        sms_queue = policy_renewal_sms(family_template)

        # then
        policy_expiring.refresh_from_db()
        policy_not_expired_yet.refresh_from_db()

        self.assertTrue(len(sms_queue) > 0)
        insuree_sms = [sms for sms in sms_queue if sms.phone == "+33644444719"]
        self.assertEquals(len(insuree_sms), 1)
        self.assertEquals(
            insuree_sms[0].sms_message, "FAMSMS;TESTCHFSMS;Test Last;Test product VISIT"
        )

        officer_sms = [sms for sms in sms_queue if sms.phone == "+32444444444"]
        self.assertEquals(len(officer_sms), 1)
        self.assertIn("TESTCHFSMS", officer_sms[0].sms_message)
        self.assertIn(family.location.name, officer_sms[0].sms_message)
        self.assertIn(family.location.parent.name, officer_sms[0].sms_message)
        self.assertIn(family.location.parent.parent.name, officer_sms[0].sms_message)
        self.assertIn("Test product VISIT", officer_sms[0].sms_message)

        # tearDown
        officer.policy_renewals.all().delete()
        policy_expiring.insuree_policies.all().delete()
        policy_expiring.delete()
        policy_not_expired_yet.insuree_policies.all().delete()
        policy_not_expired_yet.delete()
        officer.delete()
        product.delete()
        insuree.delete()
        family.delete()

    def test_insert_renewal_details(self):
        # Given
        from core import datetime, datetimedelta

        insuree_newpic, family_newpic = create_test_insuree_for_policy(
            custom_props={"photo_date": datetime.datetime.now() - datetimedelta(days=30)})
        insuree_oldpic, family_oldpic = create_test_insuree_for_policy(
            custom_props={"photo_date": "2010-01-01", "chf_id": "CHFMARK", 'last_name':'Test Last'})  # 5 years by default
        product = create_test_product("VISIT")
        officer = create_test_officer(custom_props={"phone": "+32444444444", "phone_communication": True})
        photo_newpic = create_test_photo(insuree_newpic.id, officer.id)
        photo_oldpic = create_test_photo(insuree_oldpic.id, officer.id)

        (policy_new_pic, inspolicy_new_pic) = create_test_policy2(
            product=product,
            insuree=insuree_newpic,
            custom_props={
                "expiry_date": datetime.datetime.now() + datetimedelta(days=5),
                "officer": officer,
            },
        )
        (policy_old_pic, inspolicy_old_pic) = create_test_policy2(
            product=product,
            insuree=insuree_oldpic,
            custom_props={
                "expiry_date": datetime.datetime.now() + datetimedelta(days=5),
                "officer": officer,
            },
        )

        # when
        insert_renewals(officer_id=officer.id)

        # then
        renewals_new = PolicyRenewal.objects.filter(insuree=insuree_newpic)
        expected_renewal = renewals_new.filter(policy=policy_new_pic).first()
        self.assertIsNotNone(expected_renewal)
        self.assertIsNone(expected_renewal.details.first())

        renewals_old = PolicyRenewal.objects.filter(insuree=insuree_oldpic)
        expected_renewal_old = renewals_old.filter(policy=policy_old_pic).first()
        self.assertIsNotNone(expected_renewal_old)
        detail = expected_renewal_old.details.first()
        self.assertIsNotNone(detail)

        # ALSO WHEN
        sms_queue = policy_renewal_sms("UNUSED")  # Uses the default template
        self.assertEquals(len(sms_queue), 2)
        old_sms = [sms.sms_message for sms in sms_queue if "CHFMARK" in sms.sms_message]
        self.assertEquals(len(old_sms), 1)
        self.assertTrue("HOF\nCHFMARK\nTest Last First Second\n\n" in old_sms[0])

        # tearDown
        renewals_old.first().details.all().delete()
        renewals_old.delete()
        renewals_new.first().details.all().delete()
        renewals_new.delete()
        inspolicy_old_pic.delete()
        policy_old_pic.delete()
        inspolicy_new_pic.delete()
        policy_new_pic.delete()
        officer.delete()
        product.delete()
        photo_newpic.delete()
        photo_oldpic.delete()
        insuree_oldpic.delete()
        family_oldpic.delete()
        insuree_newpic.delete()
        family_newpic.delete()



# Generated by Django 3.2.18 on 2023-05-12 08:48

import core.fields
import datetime
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('insuree', '0013_auto_20211103_1023'),
        ('core', '0019_extended_field'),
        ('product', '0006_insert_ceiling_type'),
        ('policy', '0004_add_medical_oficer_reading_rights'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='policy',
            name='row_id',
        ),
        migrations.AddField(
            model_name='policy',
            name='family',
            field=models.ForeignKey(db_column='FamilyID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='policies', to='insuree.family'),
        ),
        migrations.AddField(
            model_name='policy',
            name='officer',
            field=models.ForeignKey(blank=True, db_column='OfficerID', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='policies', to='core.officer'),
        ),
        migrations.AddField(
            model_name='policy',
            name='product',
            field=models.ForeignKey(db_column='ProdID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='policies', to='product.product'),
        ),
        migrations.AddField(
            model_name='policy',
            name='uuid',
            field=models.CharField(db_column='PolicyUUID', default=uuid.uuid4, max_length=36, unique=True),
        ),
        migrations.AddField(
            model_name='policyrenewal',
            name='insuree',
            field=models.ForeignKey(db_column='InsureeID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='policy_renewals', to='insuree.insuree'),
        ),
        migrations.AddField(
            model_name='policyrenewal',
            name='new_officer',
            field=models.ForeignKey(blank=True, db_column='NewOfficerID', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='policy_renewals', to='core.officer'),
        ),
        migrations.AddField(
            model_name='policyrenewal',
            name='new_product',
            field=models.ForeignKey(db_column='NewProdID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='policy_renewals', to='product.product'),
        ),
        migrations.AddField(
            model_name='policyrenewal',
            name='policy',
            field=models.ForeignKey(db_column='PolicyID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='policy_renewals', to='policy.policy'),
        ),
        migrations.AlterField(
            model_name='policy',
            name='validity_from',
            field=core.fields.DateTimeField(db_column='ValidityFrom', default=datetime.datetime.now),
        ),
    ]



import core

from django.test import TestCase
from insuree.test_helpers import create_test_insuree, create_test_photo
from product.test_helpers import create_test_product
from insuree.models import Relation
from policy.values import *
from policy.test_helpers import create_test_policy
from core.apps import CoreConfig
from dateutil.relativedelta import relativedelta

class PolicyValuesTestCase(TestCase):
    test_child_dob=py_datetime.date.today()- relativedelta(years= CoreConfig.age_of_majority-2)
    def test_new_policy_basis(self):
        head_insuree = create_test_insuree(with_family=True, custom_props={"dob": core.datetime.date(1985, 5, 5)})
        spouse = Relation.objects.get(id=8)
        create_test_insuree( with_family = False,
            custom_props={
                "dob": core.datetime.date(1989, 9, 9),
                "relationship": spouse,
                "family": head_insuree.family
            })

        product = create_test_product("SIMPLE", custom_props={
            "max_members": 5,
            "administration_period": 0,
            "lump_sum": 0,
            "premium_adult": 300,
            "premium_child": 200,
            "registration_lump_sum": 250,
            "general_assembly_lump_sum": 130,
            "insurance_period": 12,
        })
        policy = create_test_policy(product, head_insuree, custom_props={
            "enroll_date": core.datetime.date(2020, 11, 10),
        })
        policy, warnings = policy_values(policy, head_insuree.family, None)
        self.assertEquals(policy.start_date, core.datetime.date(2020, 11, 10))
        self.assertEquals(policy.expiry_date, core.datetime.date(2021, 11, 9))
        self.assertEquals(policy.value, 980)  # 2 x 300 + 250 + 130

        # let's add a child and shift date to 1st of a month
        child = Relation.objects.get(id=4)
        create_test_insuree(with_family = False,
            custom_props={
                "dob": self.test_child_dob,
                "relationship": child,
                "family": head_insuree.family
            })
        policy = create_test_policy(product, head_insuree, custom_props={
            "enroll_date": core.datetime.date(2020, 11, 1),
        })
        policy, warnings = policy_values(policy, head_insuree.family, None)
        self.assertEquals(policy.start_date, core.datetime.date(2020, 11, 1))
        self.assertEquals(policy.expiry_date, core.datetime.date(2021, 10, 31))
        self.assertEquals(policy.value, 1180)  # 2 x 300 + 200 + 250 + 130

    def test_new_policy_lump_sum_and_cycles(self):
        head_insuree = create_test_insuree(with_family=True, custom_props={"dob": core.datetime.date(1985, 5, 5)})
        spouse = Relation.objects.get(id=8)
        create_test_insuree( with_family = False,
            custom_props={
                "dob": core.datetime.date(1989, 9, 9),
                "relationship": spouse,
                "family": head_insuree.family
            })

        product = create_test_product("SIMPLE", custom_props={
            "max_members": 5,
            "administration_period": 0,
            "lump_sum": 200,
            "threshold": 2,
            "grace_period_enrolment": 1,
            "start_cycle_1": "01-01",
            "start_cycle_2": "01-06",
            "premium_adult": 300,
            "premium_child": 200,
            "registration_lump_sum": 0,
            "registration_fee": 10,
            "general_assembly_lump_sum": 0,
            "general_assembly_fee": 5,
            "insurance_period": 12,
        })
        policy = create_test_policy(product, head_insuree, custom_props={
            "enroll_date": core.datetime.date(2020, 11, 10),
        })
        policy, warnings = policy_values(policy, head_insuree.family, None)
        self.assertEquals(policy.start_date, core.datetime.date(2021, 1, 1))
        self.assertEquals(policy.expiry_date, core.datetime.date(2021, 12, 31))
        self.assertEquals(policy.value, 230)  # 200 + 2 x 10 + 2 x 5

        # let's add a child (outside threshold)  and shift date in 1st cycle grace period
        child = Relation.objects.get(id=4)
        create_test_insuree( with_family = False,
            custom_props={
                "dob": self.test_child_dob,
                "relationship": child,
                "family": head_insuree.family
            })
        policy = create_test_policy(product, head_insuree, custom_props={
            "enroll_date": core.datetime.date(2021, 1, 11),
        })
        policy, warnings = policy_values(policy, head_insuree.family, None)
        self.assertEquals(policy.start_date, core.datetime.date(2021, 1, 1))
        self.assertEquals(policy.expiry_date, core.datetime.date(2021, 12, 31))
        self.assertEquals(policy.value, 445)  # 200 + 1 x 200 + 3 x 10 + 3 x 5

    def test_new_policy_admin_period_max_members_insurance_period(self):
        head_insuree = create_test_insuree(with_family=True, custom_props={"chf_id":"tnpapmmip","dob": core.datetime.date(1985, 5, 5)})
        spouse = Relation.objects.get(id=8)
        create_test_insuree(with_family = False,
            custom_props={
                "dob": core.datetime.date(1989, 9, 9),
                "relationship": spouse,
                "family": head_insuree.family
            })

        product = create_test_product("SIMPLE", custom_props={
            "max_members": 2,
            "administration_period": 1,
            "lump_sum": 200,
            "threshold": 1,
            "grace_period_enrolment": 1,
            "start_cycle_1": "01-01",
            "start_cycle_2": "01-06",
            "premium_adult": 300,
            "premium_child": 200,
            "registration_lump_sum": 0,
            "registration_fee": 10,
            "general_assembly_lump_sum": 0,
            "general_assembly_fee": 5,
            "insurance_period": 6,
        })
        policy = create_test_policy(product, head_insuree, custom_props={
            "enroll_date": core.datetime.date(2021, 1, 10),
        })
        policy, warnings = policy_values(policy, head_insuree.family, None)
        self.assertEquals(policy.start_date, core.datetime.date(2021, 6, 1))  # enroll + admin outside cycle 1 + grace
        self.assertEquals(policy.expiry_date, core.datetime.date(2021, 11, 30))
        self.assertEquals(policy.value, 530)  # 200 + 300 + 2 x 10 + 2 x 5

        # let's add a child... outside max members so not counted!
        # ... and shift date to fall into grace period
        child = Relation.objects.get(id=4)
        create_test_insuree(
            custom_props={
                "dob": self.test_child_dob,
                "relationship": child,
                "family": head_insuree.family
            })
        policy = create_test_policy(product, head_insuree, custom_props={
                "enroll_date": core.datetime.date(2021, 12, 11),
            }, check = False)

        policy, warnings = policy_values(policy, head_insuree.family, None)
        self.assertEquals(policy.start_date, core.datetime.date(2022, 1, 1))  # enroll + admin in cycle 1 + grace
        self.assertEquals(policy.expiry_date, core.datetime.date(2022, 6, 30))
        self.assertEquals(policy.value, 530)  # 200 + 300 + 2 x 10 + 2 x 5



# Generated by Django 3.0.3 on 2020-10-21 08:11

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0009_mutationlog_client_mutation_details'),
        ('policy', '0002_policy_renewals'),
    ]

    operations = [
        migrations.CreateModel(
            name='PolicyMutation',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('mutation', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='policies', to='core.MutationLog')),
                ('policy', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='mutations', to='policy.Policy')),
            ],
            options={
                'db_table': 'location_PolicyMutation',
                'managed': True,
            },
        ),
        migrations.DeleteModel(
            name='PolicyRenewalDetail',
        ),
    ]
